;--------------------------
; Open evaluator for ASEL 
; possible to use meta-circularly
;------- ASEL -------------
;<exp> = <var> | <constant> | <lambda> | <if> | <definition> |
;        <assignment> | <application>
;<var>        = <scheme-symbol>
;<constant>   = <scheme-literal> | <quoted>
;<quoted>     = '<scheme-value>
;<lambda>     = (lambda <formals> <sequence>)
;<sequence>   = <exp>+
;<formals>    = () | <var> | (<var>+ [. <formals>])
;<if>         = (if <exp> <exp> <exp>)
;<definition> = (define <var> <exp>)
;<assignment> = (set! <var> <exp>)
;<application>= (<exp> <exp>*)
;--------------------
;This file is based on a copy of "open-simple.scm"
;It has been converted a bit to simulate an infinite tower.
;All lambdas have been replaced by similar lambdas with an extra first argument:
;a stack of meta-continuations
;all calls to such procedures similarly have been converted to pass on a 
;stack of meta continuations.
;Note that continuations are also procedures and thus also have to receive a
;stack of meta-continuations as first argument.
;
;The following variable names are used throughout the file
;m : stack of meta continuations
;k : continuation
;e : expression
;r : environment
;;---------
; This "infinite tower" version can be used to express fixpoint equations
;;---------

;;---- Stack of meta continuations for simulating an infinite tower

(define make-default-stack 
  (lambda (level)
    (list 'cstack level (- level 1))))

(define push 
  (lambda (base-stack cont)
    (list 'cstack cont base-stack)))

(define pop
  (lambda (stack)
    (if (number? (3rd stack))
        (make-default-stack (3rd stack))
        (3rd stack))))

(define top
  (lambda (stack)
    (if (number? (2nd stack))
      (make-loop (2nd stack))
      (2nd stack))))

;; procedure for creating the revpl procedure for default continuations
(define make-loop
  (lambda (level)
    (define loop
      (lambda (m v)
        (display level)
        (display ": ")
        (if (eq? v (void))
          "nothing is displayed"
          (write v))
        (newline)
        (display level)
        (display "> ")
        (basic-eval m (read) global-env loop)))
    loop))

;;sometimes we need something that behaves like (lambda (v) v) as continuation
(define id-cont
  (push 'should-not-be-used (lambda (m v) v)))

;---------------------

(define meta 
  (lambda (dispatch-table)

    (define evaluate
      (lambda (m e r k)
        (if (pair? e)
           (find-pair (car e) dispatch-table
              (lambda (success-pair)
                ((cdr success-pair) m evaluate e r k))
              (lambda ()
                (basic-evaluate m e r k)))
           (basic-evaluate m e r k))))

    (define basic-evaluate
      (lambda (m e r k)
        ((if (constant? e)
           evaluate-constant
           (if (variable? e)
             evaluate-variable
             (if (if? e)
               evaluate-if
               (if (assignment? e)
                 evaluate-assignment
                 (if (definition? e)
                   evaluate-definition
                   (if (abstraction? e)
                     evaluate-abstraction
                     evaluate-combination))))))
         m e r k)))

    (define evaluate-constant
      (lambda (m e r k)
        (k m (constant-part e))))

    (define evaluate-variable
      (lambda (m e r k)
        (get-pair e r
           (lambda (success-pair)
             (k m (cdr success-pair)))
           (lambda ()
             (wrong m "symbol not bound: " e)))))

    (define evaluate-if
      (lambda (m e r k)
        (evaluate m (test-part e) r
           (lambda (m v)
             (if v
               (evaluate m (then-part e) r k)
               (evaluate m (else-part e) r k))))))

    (define evaluate-assignment
      (lambda (m e r k)
        (evaluate m (value-part e) r
           (lambda (m v)
             (find-pair (id-part e) (car r)
                (lambda (success-pair)
                  (set-cdr! success-pair v)
                  (k m (void)))
                (lambda ()
                  (set-car! global-env
                       (cons (cons (id-part e) v)
                             (car global-env)))
                  (k m (void))))))))

    (define evaluate-definition
      (lambda (m e r k)
        (evaluate m (value-part e) r
           (lambda (m v)
             (find-pair (id-part e) (car r)
                (lambda (success-pair)
                  (set-cdr! success-pair v)
                  (k m (void)))
                (lambda ()
                  (set-car! r
                       (cons (cons (id-part e) v)
                             (car r)))
                  (k m (void))))))))
                
    (define evaluate-abstraction
      (lambda (m e r k)
        (k m (make-compound 
              (formals-part e) (body-part e) r))))

    (define evaluate-combination
      (lambda (m e r k)
        ;(display "@: ")
        ;(write e)
        ;(newline)
        (evaluate m (operator-part e) r
           (lambda (m proc)
             (evaluate-operands m (operands-part e) r
                (lambda (m args)
                  (apply-procedure m proc args k)))))))

    (define evaluate-operands
      (lambda (m operands r k)
        (if (null? operands)
          (k m '())
          (evaluate m (car operands) r
             (lambda (m v)
               (evaluate-operands m (cdr operands) r
                  (lambda (m w)
                    (k m (cons v w)))))))))

    (define evaluate-sequence
      (lambda (m body r k)
        (if (null? (cdr body))
          (evaluate m (car body) r k)
          (evaluate m (car body) r
            (lambda (m v)
              (evaluate-sequence m (cdr body) r k))))))

    (define make-compound
      (lambda (formals body r)
        (lambda (m . args)
          (evaluate-sequence (pop m) body (extend r formals args) (top m)))))

    ((top m) (pop m) evaluate)))

(define wrong
  (lambda (m message object)
    (display "Error:")
    (display message)
    (display object)
    (newline)
    ((top m) (pop m) 'error)))

(define apply-procedure
      (lambda (m proc args k)
        (if (procedure? proc)
          (apply proc (cons (push m k) args))
          (wrong m "operator is not a procedure" proc))))

(define extend
    (lambda (r ids vals)
      (cons (extend-frame '() ids vals) r)))

(define extend-frame
  (lambda (f ids vals)
    (if (null? ids)
      f
      (if (pair? ids)
        (extend-frame (cons (cons (car ids) (car vals)) f)
                      (cdr ids)
                        (cdr vals))
        (cons (cons ids vals) f)))))

(define get-pair
    (lambda (id r success failure)
      (if (null? r)
        (failure)
        (find-pair id (car r)
           success
           (lambda ()
             (get-pair id (cdr r) success failure))))))

(define find-pair
   (lambda (elt alist success failure)
     ((lambda (assq-result)
        (if assq-result
          (success assq-result)
          (failure)))
      (assq elt alist))))

(define empty-env '())

(define 1st (lambda (l) (car l)))
(define 2nd (lambda (l) (car (cdr l))))
(define 3rd (lambda (l) (car (cdr (cdr l)))))
(define 4th (lambda (l) (car (cdr (cdr (cdr l))))))
(define 5th (lambda (l) (car (cdr (cdr (cdr (cdr l)))))))

(define test-tag
  (lambda (tag)
    (lambda (e)
      (if (pair? e) (eq? (car e) tag) #f))))

(define make-primitive ;;use only for "non-higher order" primitives
  (lambda (op)
    (lambda (m . args)
      ((top m) (pop m) (apply op args)))))

(define primitive-identifiers
   (lambda ()
    '(car cdr cons set-car! set-cdr! assq memq
      null? = eq? newline write display read
      + - * symbol? list pair? eof-object?
      close-input-port open-input-file void procedure?)))

(define primitive-procs
  (lambda ()
    (list car cdr cons set-car! set-cdr! assq memq
      null? = eq? newline write display read
      + - * symbol? list pair? eof-object?
      close-input-port open-input-file void procedure?)))

(define variable? symbol?)
(define if? (test-tag 'if))
(define assignment? (test-tag 'set!))
(define definition? (test-tag 'define))
(define abstraction? (test-tag 'lambda))
(define quote? (test-tag 'quote))

(define constant?
  (lambda (e)
    (if (pair? e) (quote? e)
      (if (symbol? e) #f #t))))

(define constant-part
    (lambda (e) (if (quote? e) (2nd e) e)))

(define test-part 2nd)
(define then-part 3rd)
(define else-part 4th)

(define id-part 2nd)
(define value-part 3rd)

(define formals-part 2nd)
(define body-part (lambda (e) (cdr (cdr e))))

(define operator-part 1st)
(define operands-part cdr)

(define void
    ((lambda (v) (lambda () v)) (cons '* '*)))

(define mapper
      (lambda (f l)
        (if (null? l)
          '()
          (cons (f (car l)) (mapper f (cdr l))))))

(define openloop
  (lambda (m evaluate read-prompt write-prompt)
    (display read-prompt)
    (evaluate m (read) global-env
      (lambda (m v)
        (display write-prompt)
        (if (eq? v (void))
          "Nothing is displayed"
          (write v))
        (newline)
        (openloop m evaluate read-prompt write-prompt)))))

(define initialize-global-env
      (lambda ()
        (set! global-env
           (extend
             empty-env
             (cons 'apply (primitive-identifiers))
             (cons (lambda (m proc args)
                     (apply-procedure (pop m) proc args (top m)))
                   (mapper make-primitive (primitive-procs)))))
        (set! global-env
              (extend global-env 
                      '(meta* default-dispatcher openloop)
                  (list meta  default-dispatcher openloop)))))

(define default-dispatcher '())

(define basic-eval (meta id-cont default-dispatcher))
    
(define start
  (lambda ()
    (initialize-global-env)
    (set-car! global-env (cons (cons 'global-env global-env) (car global-env)))
    (let ((s (make-default-stack 0)))
      ((top s) (pop s) 'begin))))


(define global-env 'dummy) ;;just so that global-env exists and can be set! to
