;--------------------------
; Open evaluator for ASEL 
; possible to use meta-circularly
;------- ASEL -------------
;<exp> = <var> | <constant> | <lambda> | <if> | <definition> |
;        <assignment> | <application>
;<var>        = <scheme-symbol>
;<constant>   = <scheme-literal> | <quoted>
;<quoted>     = '<scheme-value>
;<lambda>     = (lambda <formals> <sequence>)
;<sequence>   = <exp>+
;<formals>    = () | <var> | (<var>+ [. <formals>])
;<if>         = (if <exp> <exp> <exp>)
;<definition> = (define <var> <exp>)
;<assignment> = (set! <var> <exp>)
;<application>= (<exp> <exp>*)
;--------------------
;This implementation fully abstracts procedures, they are "black boxes". The
;only thing the user can do with procedures is apply them to some arguments
;Procedures are represented by procedures (of the meta-language) 
;of the form
;(lambda (k . args) ...) 
;That is a procedure is represented by procedure that takes the represendted
;procedure's arguments augmented with an extra first argument, the continuation
;--------------------
;This open implementation is "meta-scircular" and can be used to build 
;"literal" towers of open implementations.
;--------------------

(define meta 
  (lambda (dispatch-table)

    (define evaluate
      (lambda (e r k)
        (if (pair? e)
           (find-pair (car e) dispatch-table
              (lambda (success-pair)
                ((cdr success-pair) evaluate e r k))
              (lambda ()
                (basic-evaluate e r k)))
           (basic-evaluate e r k))))

    (define basic-evaluate
      (lambda (e r k)
        ((if (constant? e)
           evaluate-constant
           (if (variable? e)
             evaluate-variable
             (if (if? e)
               evaluate-if
               (if (assignment? e)
                 evaluate-assignment
                 (if (definition? e)
                   evaluate-definition
                   (if (abstraction? e)
                     evaluate-abstraction
                     evaluate-combination))))))
         e r k)))

    (define evaluate-constant
      (lambda (e r k)
        (k (constant-part e))))

    (define evaluate-variable
      (lambda (e r k)
        (get-pair e r
           (lambda (success-pair)
             (k (cdr success-pair)))
           (lambda ()
             (wrong "symbol not bound: " e)))))

    (define wrong
      (lambda (message object)
        (display "Error:")
        (display message)
        (display object)
        (newline)))

    (define evaluate-if
      (lambda (e r k)
        (evaluate (test-part e) r
           (lambda (v)
             (if v
               (evaluate (then-part e) r k)
               (evaluate (else-part e) r k))))))

    (define evaluate-assignment
      (lambda (e r k)
        (evaluate (value-part e) r
           (lambda (v)
             (find-pair (id-part e) (car r)
                (lambda (success-pair)
                  (set-cdr! success-pair v)
                  (k (void)))
                (lambda ()
                  (set-car! global-env
                       (cons (cons (id-part e) v)
                             (car global-env)))
                  (k (void))))))))

    (define evaluate-definition
      (lambda (e r k)
        (evaluate (value-part e) r
           (lambda (v)
             (find-pair (id-part e) (car r)
                (lambda (success-pair)
                  (set-cdr! success-pair v)
                  (k (void)))
                (lambda ()
                  (set-car! r
                       (cons (cons (id-part e) v)
                             (car r)))
                  (k (void))))))))
                
    (define evaluate-abstraction
      (lambda (e r k)
        (k (make-compound 
            (formals-part e) (body-part e) r))))

    (define evaluate-combination
      (lambda (e r k)
        ;(display "@: ")
        ;(write e)
        ;(newline)
        (evaluate (operator-part e) r
           (lambda (proc)
             (evaluate-operands (operands-part e) r
                (lambda (args)
                  (apply-procedure proc args k)))))))

    (define evaluate-operands
      (lambda (operands r k)
        (if (null? operands)
          (k '())
          (evaluate (car operands) r
             (lambda (v)
               (evaluate-operands (cdr operands) r
                  (lambda (w)
                    (k (cons v w)))))))))

    (define evaluate-sequence
      (lambda (body r k)
        (if (null? (cdr body))
          (evaluate (car body) r k)
          (evaluate (car body) r
            (lambda (v)
              (evaluate-sequence (cdr body) r k))))))

    (define make-compound
      (lambda (formals body r)
        (lambda (k . args)
          (evaluate-sequence body (extend r formals args) k))))

    evaluate))

(define apply-procedure
      (lambda (proc args k)
        (if (procedure? proc)
          (apply proc (cons k args))
          (wrong "operator is not a procedure" proc))))

(define extend
    (lambda (r ids vals)
      (cons (extend-frame '() ids vals) r)))

(define extend-frame
  (lambda (f ids vals)
    (if (null? ids)
      f
      (if (pair? ids)
        (extend-frame (cons (cons (car ids) (car vals)) f)
                      (cdr ids)
                        (cdr vals))
        (cons (cons ids vals) f)))))

(define get-pair
    (lambda (id r success failure)
      (if (null? r)
        (failure)
        (find-pair id (car r)
           success
           (lambda ()
             (get-pair id (cdr r) success failure))))))

(define find-pair
   (lambda (elt alist success failure)
     ((lambda (assq-result)
        (if assq-result
          (success assq-result)
          (failure)))
      (assq elt alist))))

(define empty-env '())

(define 1st (lambda (l) (car l)))
(define 2nd (lambda (l) (car (cdr l))))
(define 3rd (lambda (l) (car (cdr (cdr l)))))
(define 4th (lambda (l) (car (cdr (cdr (cdr l))))))
(define 5th (lambda (l) (car (cdr (cdr (cdr (cdr l)))))))

(define test-tag
  (lambda (tag)
    (lambda (e)
      (if (pair? e) (eq? (car e) tag) #f))))

(define make-primitive ;;only for "non-higher order" primitives
  (lambda (op)
    (lambda (k . args)
      (k (apply op args)))))

(define primitive-identifiers
   (lambda ()
    '(car cdr cons set-car! set-cdr! assq memq
      null? = eq? newline write display read
      + - * symbol? list pair? eof-object?
      close-input-port open-input-file void procedure?)))

(define primitive-procs
  (lambda ()
    (list car cdr cons set-car! set-cdr! assq memq
      null? = eq? newline write display read
      + - * symbol? list pair? eof-object?
      close-input-port open-input-file void procedure?)))

(define variable? symbol?)
(define if? (test-tag 'if))
(define assignment? (test-tag 'set!))
(define definition? (test-tag 'define))
(define abstraction? (test-tag 'lambda))
(define quote? (test-tag 'quote))

(define constant?
  (lambda (e)
    (if (pair? e) (quote? e)
      (if (symbol? e) #f #t))))

(define constant-part
    (lambda (e) (if (quote? e) (2nd e) e)))

(define test-part 2nd)
(define then-part 3rd)
(define else-part 4th)

(define id-part 2nd)
(define value-part 3rd)

(define formals-part 2nd)
(define body-part (lambda (e) (cdr (cdr e))))

(define operator-part 1st)
(define operands-part cdr)

(define void
    ((lambda (v) (lambda () v)) (cons '* '*)))

(define mapper
      (lambda (f l)
        (if (null? l)
          '()
          (cons (f (car l)) (mapper f (cdr l))))))

(define initialize-global-env
      (lambda ()
        (set! global-env
           (extend
             empty-env
             (cons 'apply (primitive-identifiers))
             (cons (lambda (k proc args)
                     (apply-procedure proc args k))
                   (mapper make-primitive (primitive-procs)))))))

(define default-dispatcher '())

(define basic-eval (meta default-dispatcher))
    
(define openloop
  (lambda (evaluate read-prompt write-prompt)
    (display read-prompt)
    (evaluate (read) global-env (repl-cont evaluate read-prompt write-prompt))))

(define repl-cont
  (lambda (evaluate read-prompt write-prompt)
    (lambda (v)
      (display write-prompt)
      (if (eq? v (void))
        "Nothing is displayed"
        (write v))
      (newline)
      (openloop evaluate read-prompt write-prompt))))

(define loadfile
  (lambda (evaluate file)
    ((lambda (port)
       ((lambda (loop)
          (set! loop
            (lambda (v)
              (if (eof-object? v)
                (close-input-port port)
                (evaluate v global-env
                   (lambda (ignore)
                     (loop (read port)))))))
          (loop (read port)))
        '*))
     (open-input-file file))))

(define boot-1-level 
  (lambda (evaluate input-prompt output-prompt)
    (initialize-global-env)
    (loadfile basic-eval this-file-name)
    (openloop evaluate input-prompt output-prompt)))

(define start
  (lambda (evaluate input-prompt output-prompt)
    (initialize-global-env)
    (openloop evaluate input-prompt output-prompt)))
   
(define global-env 'dummy) ;;just so that global-env exists and can be set! to

(define this-file-name "open-simple.scm")